"""Extended Textual Sparkline widget."""

from typing import Any
from typing import Callable
from typing import Sequence

from textual import events
from textual import widgets

from .._extensions import WidgetExtension


class Sparkline(WidgetExtension, widgets.Sparkline):
    """An extended sparkline widget to display numerical data."""

    def __init__(
        self,
        data: Sequence[float] | None = None,
        *,
        summary_function: Callable[[Sequence[float]], float] | None = None,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        styles: dict[str, Any] | None = None,
        disabled_messages: list[type[events.Message]] | None = None,
        callbacks: dict[str, Callable] | None = None,
    ) -> None:
        """Initialize a sparkline widget.

        Args:
            data: The initial data to populate the sparkline with.
            summary_function: Summarises bar values into a single value used to
                represent each bar.
            name: The name of the widget.
            id: The ID of the widget in the DOM.
            classes: The CSS classes for the widget.
            disabled: Whether the widget is disabled or not.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        super().__init__(
            data,
            summary_function=summary_function,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self.__extend_widget__(
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )



"""Extended Textual checkbox widget."""

from textual import widgets

from .._extensions import ToggleButtonInitExtension


class Checkbox(ToggleButtonInitExtension, widgets.Checkbox):  # pylint: disable=too-many-ancestors
    """An extended checkbox widget that represents a boolean value."""



"""Extended Textual tooltip widget."""

from textual import widgets

from .._extensions import StaticInitExtension


class Tooltip(StaticInitExtension, widgets.Tooltip):
    """An extended widget for displaying short tip messages."""



"""Extended Textual radio button widget."""

from textual import widgets

from .._extensions import ToggleButtonInitExtension


class RadioButton(ToggleButtonInitExtension, widgets.RadioButton):  # pylint: disable=too-many-ancestors
    """An extended radio button widget that represents a boolean value."""



"""Extended Textual RadioSet widget."""

from typing import Any
from typing import Callable

from textual import events
from textual import widgets

from .._extensions import WidgetExtension


class RadioSet(WidgetExtension, widgets.RadioSet):
    """Widget for grouping a collection of radio buttons into a set."""

    def __init__(
        self,
        *buttons: str | widgets.RadioButton,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        styles: dict[str, Any] | None = None,
        disabled_messages: list[type[events.Message]] | None = None,
        callbacks: dict[str, Callable] | None = None,
    ) -> None:
        """Initialize the radio set.

        Args:
            buttons: A collection of labels or "RadioButton"(s) to group together.
            name: The name of the radio set.
            id: The ID of the radio set in the DOM.
            classes: The CSS classes of the radio set.
            disabled: Whether the radio set is disabled or not.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        super().__init__(
            *buttons,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self.__extend_widget__(
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )



"""Extended Textual label widget for displaying text-oriented renderables."""

from textual import widgets

from .._extensions import StaticInitExtension


class Label(StaticInitExtension, widgets.Label):
    """An extended, simple, label widget for displaying text-oriented renderables."""



"""Extended Textual container widgets."""

from __future__ import annotations

from textual import containers

from .._extensions import WidgetInitExtension


class Center(WidgetInitExtension, containers.Center):
    """An extended container which aligns children on the X axis."""


class Container(WidgetInitExtension, containers.Container):
    """An extended, simple container widget, with vertical layout."""


class Grid(WidgetInitExtension, containers.Grid):
    """An extended container with grid layout."""


class Horizontal(WidgetInitExtension, containers.Horizontal):
    """An extended container with horizontal layout and no scrollbars."""


class HorizontalScroll(WidgetInitExtension, containers.HorizontalScroll):
    """An extended container with horizontal layout and an automatic scrollbar on the Y axis."""


class Middle(WidgetInitExtension, containers.Middle):
    """An extended container which aligns children on the Y axis."""


class PageContainer(Container):
    """Container used to display page contents, and signal to multi-page applications which ID to use in callbacks.

    If not used in a multi-page app, it is functionally the same as a Container.
    """


class ScrollableContainer(containers.ScrollableContainer):
    """An extended scrollable container with vertical layout, and auto scrollbars on both axis."""


class Vertical(WidgetInitExtension, containers.Vertical):
    """An extended container with vertical layout and no scrollbars."""


class VerticalScroll(WidgetInitExtension, containers.VerticalScroll):
    """An extended container with vertical layout and an automatic scrollbar on the Y axis."""



"""Extended Textual Tab Content widgets."""

from typing import Any
from typing import Callable

from rich.text import TextType
from textual import events
from textual import widgets
from textual.widget import Widget

from .._extensions import WidgetExtension


class TabbedContent(WidgetExtension, widgets.TabbedContent):
    """An extended container with associated tabs to toggle content visibility."""

    def __init__(
        self,
        *titles: TextType,
        initial: str = "",
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        styles: dict[str, Any] | None = None,
        disabled_messages: list[type[events.Message]] | None = None,
        callbacks: dict[str, Callable] | None = None,
    ) -> None:
        """Initialize a TabbedContent widgets.

        Args:
            *titles: Positional argument will be used as title.
            initial: The id of the initial tab, or empty string to select the first tab.
            name: The name of the button.
            id: The ID of the button in the DOM.
            classes: The CSS classes of the button.
            disabled: Whether the button is disabled or not.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        super().__init__(
            *titles,
            initial=initial,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self.__extend_widget__(
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )


class TabPane(WidgetExtension, widgets.TabPane):
    """An extended container for switchable content, with additional title."""

    def __init__(
        self,
        title: TextType,
        *children: Widget,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        styles: dict[str, Any] | None = None,
        disabled_messages: list[type[events.Message]] | None = None,
        callbacks: dict[str, Callable] | None = None,
    ) -> None:
        """Initialize a TabPane.

        Args:
            title: Title of the TabPane (will be displayed in a tab label).
            *children: Widget to go inside the TabPane.
            name: Optional name for the TabPane.
            id: Optional ID for the TabPane.
            classes: Optional initial classes for the widget.
            disabled: Whether the TabPane is disabled or not.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        super().__init__(
            title,
            *children,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self.__extend_widget__(
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )



"""Common utilities for interacting with the base textual module."""

from collections import namedtuple

from textual import __version__ as __lib_version__  # Dynamic attribute in textual. pylint: disable=no-name-in-module

TextualVersion = namedtuple("TextualVersion", ["major", "minor", "maintenance"])
major, minor, maintenance = __lib_version__.split(".")
major, minor, maintenance = int(major), int(minor), int(maintenance)

textual_version = TextualVersion(major, minor, maintenance)



"""Widget container for basic data storage and sharing between callbacks."""

import time
from typing import Any
from typing import Callable
from typing import TypeVar

from textual import events
from textual.reactive import reactive
from textual.widget import Widget

from ._extensions import WidgetInitExtension

JsonType = TypeVar("JsonType", dict, list, bool, float, int, str, None)  # Custom type. pylint: disable=invalid-name


class Store(WidgetInitExtension, Widget):
    """Hidden widget for basic data storage.

    Intended to only be used for sharing data between callbacks.
    """

    DEFAULT_CSS = """
    Store {
        display: none;
        width: 0;
        height: 0;
    }
    """

    # The currently stored data.
    data: JsonType = reactive(None, repaint=False, init=False)
    # The timestamp from the last time the value was modified.
    modified_timestamp: float = reactive(-1.0, repaint=False, init=False)
    # Sentinel value used to trigger clears from callbacks. Set to True to manually trigger a clear.
    clear_data: bool = reactive(False, always_update=True, repaint=False, init=False)

    def __init__(
        self,
        data: Any = None,
        id: str | None = None,
        disabled_messages: list[events.Message] | None = None,
        callbacks: dict[str, Callable] | None = None,
    ) -> None:
        """Initialize the data store.

        Args:
            data: Initial data to store.
            id: The ID of the widget in the DOM.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        super().__init__(id=id, disabled_messages=disabled_messages, callbacks=callbacks)
        self.data = data

    def watch_clear_data(self, new_value: bool) -> None:
        """Monitor the sentinel attribute for clear data requests from callbacks.

        Args:
            new_value: New clear data value set on the reactive attribute.
        """
        if new_value:
            self.data = None

    def watch_data(self, _: JsonType) -> None:
        """Monitor the data in order to update the modified timestamp."""
        self.modified_timestamp = time.time()



"""Logging utilities."""

import logging
from typing import Any


class NullLogger(logging.Logger):
    """Dummy logger to allow null routing all logging messages."""

    def __getattribute__(self, name: str) -> Any:
        """Null route all attribute requests."""
        if name != "_null_route":
            return self._null_route
        return super().__getattribute__(name)

    def _null_route(self, *args: Any, **kwargs: Any) -> None:
        """Null route any callable action."""



"""Custom textual Apps with textology extensions included."""

from __future__ import annotations

import logging
from collections import defaultdict
from inspect import isawaitable
from types import ModuleType
from typing import Any
from typing import Callable

from textual import events
from textual._path import _make_path_object_relative  # pylint: disable=protected-access
from textual.app import App
from textual.app import ComposeResult
from textual.app import CSSPathType
from textual.css.query import NoMatches
from textual.driver import Driver
from textual.screen import Screen
from textual.widget import Widget

from .observers import Modified
from .observers import ObserverManager
from .observers import Select
from .observers import Update
from .pages import _GLOBAL_PAGE_MAP
from .pages import Page
from .pages import register_page
from .router import Endpoint
from .router import Request
from .widgets import Container
from .widgets import Label
from .widgets import Location
from .widgets import PageContainer
from .widgets import walk_all_children

_DEFAULT_CONTENT_ID = "content"
_DEFAULT_URL_ID = "url"


class WidgetApp(App):
    """Application with a single widget for the root layout, and basic extensions.

    Extensions:
        - Compose layout without subclassing.
        - Ability to swap themes.
    """

    CSS_THEMES: dict[str, list[CSSPathType]] | None = None
    """Mappings of additional file paths to load CSS from in addition to base CSS."""

    def __init__(
        self,
        child: Callable | Widget | None = None,
        driver_class: type[Driver] | None = None,
        css_path: CSSPathType | None = None,
        watch_css: bool = False,
        css_theme: str | list[str] | None = None,
        css_themes: dict[str, list[CSSPathType]] | None = None,
    ) -> None:
        """Initialize an application with a widget for the layout.

        Args:
            child: Root widget, or function to create root widget.
                Use a callable to delay creation until end of initialization.
                Defaults to a blank Container with "content" id.
            driver_class: Driver class or `None` to auto-detect.
                This will be used by some Textual tools.
            css_path: Path to CSS or `None` to use the `CSS_PATH` class variable.
                To load multiple CSS files, pass a list of strings or paths which will be loaded in order.
            watch_css: Reload CSS if the files changed.
                This is set automatically if you are using `textual run` with the `dev` switch.
            css_theme: Initial CSS theme to load from "css_themes".
                Themes are applied in addition to base "css_path" values, rather than in place of.
            css_themes: Mapping of CSS paths by string names, or `None` to use the `CSS_THEMES` class variable.

        Raises:
            CssPathError: When the supplied CSS path(s) are an unexpected type.
        """
        css_path = css_path or self.CSS_PATH
        if css_path and not isinstance(css_path, list):
            css_path = [css_path]
        if css_theme and not isinstance(css_theme, list):
            css_theme = [css_theme]
        self._css_theme: list[str] = css_theme
        self.css_themes = css_themes or self.CSS_THEMES or {}
        for css_theme_name, css_theme_paths in self.css_themes.items():
            self.css_themes[css_theme_name] = [
                _make_path_object_relative(css_theme_path, self) for css_theme_path in css_theme_paths
            ]
        css_path = css_path or []
        for theme in self._css_theme or []:
            if theme in self.css_themes:
                css_path.extend(self.css_themes[theme])
        super().__init__(
            driver_class=driver_class,
            css_path=css_path,
            watch_css=watch_css,
        )
        if not child:
            child = Container(id=_DEFAULT_CONTENT_ID)
        else:
            child = child() if isinstance(child, Callable) else child
        self.child = child

    def apply_theme(self, theme: str | list[str] | None = None) -> None:
        """Load a CSS theme.

        Themes are applied in addition to base "css_path" values, rather than in place of.

        Args:
            theme: Name(s) of the CSS theme(s) to load from "css_themes".
                All other stylesheets not in the base CSS configuration will be unloaded.
        """
        theme = [theme] if theme and not isinstance(theme, list) else theme
        css_paths = self.css_path
        stylesheet = self.stylesheet.copy()

        # Cleanup old themes first.
        for css_theme in self._css_theme or []:
            if css_theme not in self.css_themes:
                continue
            for css_theme_path in self.css_themes[css_theme]:
                if css_theme_path in css_paths:
                    css_paths.remove(css_theme_path)
                str_path = str(css_theme_path)
                if (str_path, "") in stylesheet.source:
                    stylesheet.source.pop((str_path, ""))
            self.log.info(f"Removed CSS theme: {self.css_theme}")

        # Apply new themes second.
        for css_theme in theme or []:
            if css_theme not in self.css_themes:
                self.log.error(f"CSS Theme not found: {css_theme}")
                continue
            for css_theme_path in self.css_themes.get(css_theme, []):
                css_paths.append(css_theme_path)

        # Force reload of CSS and layouts last.
        self.css_path = css_paths
        try:
            stylesheet.read_all(css_paths)
            stylesheet.parse()
            self.log.info(f"Applied theme: {theme}")
        except Exception as error:  # Do not crash app on theme swap failure. pylint: disable=broad-exception-caught
            self._css_has_errors = True
            self.log.error(error)
        else:
            self._css_theme = theme
            self._css_has_errors = False
            self.stylesheet = stylesheet
            self.stylesheet.update(self)
            self.refresh_css()
            self.refresh(layout=True)
            for screen in self.screen_stack:
                # Must call private refresh or rendered layout may be incorrect until next screen change.
                screen._refresh_layout(self.size, full=True)  # pylint: disable=protected-access
                self.stylesheet.update(self.screen)
                self.screen.refresh(layout=True)

    @property
    def css_theme(self) -> list[str] | None:
        """Provide the CSS themes currently applied on top of the base CSS."""
        return self._css_theme

    def compose(self) -> ComposeResult:
        """Provide the root widget for the application.

        Yields:
            Child widget set on instantiation that will be used as the root of the application.
        """
        yield self.child


class ExtendedApp(WidgetApp, ObserverManager):
    """Textual application with multiple Textology extensions for automating UI updates.

    Extensions:
        - Automatic input/output callbacks for managing application state via ".when()" registration.
        - URL based multi-page content via ".register_page()".
        - URL routing for resource requests within the application via ".location.get()".
        - URL history for navigating via ".back()", ".forward()", etc.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        child: Callable | Widget | None = None,
        use_pages: bool = False,
        pages: list[Page | ModuleType | str | Callable] | None = None,
        driver_class: type[Driver] | None = None,
        css_path: CSSPathType | None = None,
        watch_css: bool = False,
        css_theme: str | list[str] | None = None,
        css_themes: dict[str, list[CSSPathType]] | None = None,
        logger: logging.Logger | None = None,
    ) -> None:
        """Initialize an application with tracking for input/output callbacks.

        Args:
            child: Root widget, or function to create root widget.
                Use a callable to delay creation until end of initialization.
                Defaults to a blank Container with "content" id.
            use_pages: Whether to enable multi-page application support.
                When enabled, this will include automatic URL routing callbacks via "widgets.Location".
                Force enabled if "pages" are provided. Enabling without "pages" allows registering pages later.
            pages: Initial pages to load into multi-page applications.
                Refer to "register_page()" for options.
            driver_class: Driver class or `None` to auto-detect.
                This will be used by some Textual tools.
            css_path: Path to CSS or `None` to use the `CSS_PATH` class variable.
                To load multiple CSS files, pass a list of strings or paths which will be loaded in order.
            watch_css: Reload CSS if the files changed.
                This is set automatically if you are using `textual run` with the `dev` switch.
            css_theme: Initial CSS theme to load from "css_themes".
                Themes are applied in addition to base "css_path" values, rather than in place of.
            css_themes: Mapping of CSS paths by string names, or `None` to use the `CSS_THEMES` class variable.
            logger: Custom logger to send callback messages to.

        Raises:
            CssPathError: When the supplied CSS path(s) are an unexpected type.
        """
        child = child or Container(
            Location(id=_DEFAULT_URL_ID),
            Container(id=_DEFAULT_CONTENT_ID) if not use_pages else PageContainer(id=_DEFAULT_CONTENT_ID),
        )
        super().__init__(
            child=child,
            driver_class=driver_class,
            css_path=css_path,
            watch_css=watch_css,
            css_theme=css_theme,
            css_themes=css_themes,
        )
        # Manually set up observer manager mixin since App inheritance does not automatically trigger.
        ObserverManager.__init__(self, logger=logger or logging.root)
        self.attach_to_observers(self)

        self._observer_message_handler_map = {}
        self._location: Location | None = None
        self._page_registry: dict[str, Page] = {}
        self._use_pages = use_pages or bool(pages)

        self.enable_pages()
        if pages:
            for page in pages:
                self.register_page(page)

    def apply_update(
        self,
        observer_id: str,
        component: Widget,
        component_id: str,
        component_property: str,
        value: Any,
    ) -> None:
        """Apply an update operation on a Widget."""
        if component_property == "children":
            # Children is a special property on widgets, it cannot be directly applied. Manually swap children.
            component.remove_children()
            if value:
                if not isinstance(value, list):
                    value = [value]
                component.mount_all(value)
        elif component_property == "screen":
            # Screen is a special property on applications, it cannot be directly applied. Manually add to stack.
            component.app.push_screen(value)
        else:
            super().apply_update(observer_id, component, component_id, component_property, value)

    def back(self) -> int:
        """Go back one URL in the history.

        Returns:
            The new index in the history.
        """
        return self.location.back()

    @property
    def document(self) -> Screen | None:
        """Provide the base screen of the application, representing the main visible "document" in the window."""
        screens = list(self.screen_stack)
        document = screens[0] if screens else None
        return document

    def enable_pages(self) -> None:
        """Set up multi-page application routing."""
        if not self._use_pages:
            return

        location = self.location
        if not location:
            raise ValueError("Layout must contain a Location widget if pages are enabled")
        if not location.id:
            raise ValueError("Location widget must have an id if pages are enabled")

        page_container = None
        for node in walk_all_children(self.child):
            if isinstance(node, PageContainer):
                page_container = node
                break
        if page_container is None:
            raise ValueError("Layout must contain a PageContainer widget if pages are enabled")
        if not page_container.id:
            raise ValueError("PageContainer widget must have an id if pages are enabled")

        # Finish final set up after all validations are performed to ensure page routing is supported.
        self.when(
            Modified(location.id, "pathname"),
            Select(location.id, "search"),
            Update(page_container.id, "children"),
        )(self._page_router)
        self.location.endpoint_not_found = Endpoint([], "", self._page_not_found)
        for page in _GLOBAL_PAGE_MAP.values():
            self.register_page(page)

    def forward(self) -> int:
        """Go forward one URL in the history.

        Returns:
            The new index in the history.
        """
        return self.location.forward()

    def get_component(self, component_id: str) -> Any:
        """Fina a component in the DOM."""
        try:
            return self.query_one(f"#{component_id}")
        except NoMatches:
            return None

    @property
    def location(self) -> Location | None:
        """Find the application's primary Location widget for routing addresses."""
        if self._location is None:
            self._update_location()
        return self._location

    async def _on_message(self, message: events.Message) -> None:
        """Process messages after sending to registered observers first."""
        control = message.control
        if control:
            controller_id = control.id
            if controller_id:
                cls = message.__class__
                callbacks = self._observer_message_handler_map.get(f"{controller_id}@{cls.__module__}.{cls.__name__}")
                if callbacks:
                    for callback in callbacks:
                        result = callback(None, message)
                        if isawaitable(result):
                            await result
        await super()._on_message(message)

    def on_mount(self, _: events.Mount) -> None:
        """Ensure the widget and observers are fully loaded after mounting."""
        if self.logger == logging.root:
            self.logger = self.log

        # Combine global observers with local observers outside init, so that apps can be declared anywhere in modules.
        combined_observer_map = defaultdict(lambda: defaultdict(list))
        for observer_map in (self._observer_map, self._observer_map_global):
            for widget_id, property_map in observer_map.items():
                for property_id, observers in property_map.items():
                    combined_observer_map[widget_id][property_id].extend(observers)

        # Register event watchers.
        for widget_id, property_map in combined_observer_map.items():
            for property_id, observers in property_map.items():
                for observer in observers:
                    if observer.publications:
                        self._observer_message_handler_map.setdefault(f"{widget_id}@{property_id}", []).append(
                            self._generate_callback(
                                widget_id,
                                property_id,
                                observer,
                            )
                        )

    def _page_not_found(
        self,
        request: Request,
    ) -> Label:
        """Default handler for when a request is made to a page that is not available."""
        self.logger.warning(f"Page not found {request.url.path}")
        return Label("Page not found")

    @property
    def page_registry(self) -> dict[str, Page]:
        """Provide the combined page registry in use by this multi-page application."""
        return self._page_registry.copy()

    def _page_router(self, pathname: str, search: str) -> list[Widget]:
        """Load the appropriate page based on the URL path and search options."""
        self.logger.debug(f"Routing page content for: {pathname}")

        # Simulate router "serve()" to allow error handling at the callback level instead of router/widget level.
        request = Request(pathname if not search else f"{pathname}?{search}")
        path = request.url.path
        endpoint = self.location.endpoint(path, "GET")

        # Create kwargs for the layout function based off of:
        #   - App if requested, to avoid circular imports if needed.
        #   - Original request object if requested by name, to allow full access to URL values.
        #   - Inline path variables if path is a dynamic route.
        #   - User search/query variables.
        kwargs = {}
        if "app" in endpoint.handler_vars:
            kwargs["app"] = self
        if "request" in endpoint.handler_vars:
            kwargs["request"] = request
        if not endpoint.route.static:
            kwargs.update(**endpoint.route.match_groups(path))
        kwargs.update(request.query)

        return endpoint.handler(**kwargs)

    def _register_reactive_observers(self, widget: Widget) -> None:
        """Enable observers for a newly added widget and its reactive attributes if it has an ID."""
        # Do not watch any widgets or properties for changes that have not had an observer set up.
        widget_id = widget.id
        if not widget_id or (widget_id not in self._observer_map and widget_id not in self._observer_map_global):
            return
        for property_name in widget._reactives.keys():  # pylint: disable=protected-access
            if (
                property_name not in self._observer_map[widget_id]
                and property_name not in self._observer_map_global[widget_id]
            ):
                continue
            for callback in self.generate_callbacks(widget_id, property_name):
                self.watch(widget, property_name, callback, init=False)

    def register_page(  # pylint: disable=too-many-arguments
        self,
        page: Page | ModuleType | str | Callable | None = None,
        path: str | None = None,
        name: str | None = None,
        order: int = 0,
        redirect_from: str | list[str] | None = None,
        layout: Callable | None = None,
    ) -> None:
        """Register a URL path to a layout in this multi-page application.

        Args:
            page: A module, or module path, where the remaining page's variables are defined.
                e.g. If calling from within the module itself: "__name__"
                e.g. If calling from another module: "mylib.home"
            path: URL Path, with or without variables. e.g. "/", "/home", "/documents/{document_name}"
                Inferred from the "module" or "layout" if not provided.
                    e.g. "mylib.home" -> "/home"
                    e.g. "layout_home_page" -> "/home_page"
                Variables marked as {variable_name} in paths will be passed to "layout" as keyword arguments.
            name: The name of the page link, such as what might be shown in navigation menus.
                Inferred from the "path" if not provided.
                    e.g. "home_page" -> "Home Page"
            order: The relative order to sort pages in the "page_registry", such as ordering in navigation menus.
            redirect_from: Paths that should redirect to this page's path. e.g. "/v1/home"
            layout: Function to call to generate the widget(s) used in the page's layout.
        """
        if not self._use_pages:
            raise ValueError("Pages are not enabled on this application")
        page = register_page(
            page=page,
            path=path,
            name=name,
            order=order,
            redirect_from=redirect_from,
            layout=layout,
            page_map=self._page_registry,
        )
        if page.path in ("/404", "/not_found", "/not_found_404"):
            # This is a special page that is not directly routed; it is used on every invalid path request.
            self.location.endpoint_not_found = Endpoint([], "", page.layout)
        else:
            self.route(page.path)(page.layout)

    def reload(self) -> None:
        """Reload the most recent URL in the history."""
        self.location.reload()

    def route(self, path: str, methods: list[str] = ("GET",)) -> Callable:
        """Create a decorator that will register a path/method combination to a request callback.

        Args:
            path: Resource location that the decorated function will be allowed to respond to.
            methods: One or more methods that the decorated function will be allowed to respond to at the given path.

        Returns:
            A decorator that will register a function as capable of accepting requests to a specific path/method combo.
        """
        return self.location.route(path, methods=methods)

    def _update_location(self) -> None:
        """Walk the layout tree to allow finding the application's Location widget at any point in the lifecycle."""
        for node in walk_all_children(self.child):
            if isinstance(node, Location):
                self._location = node
                return


def _post_mount_patch(self: Widget) -> None:
    """Called after the object has been mounted, regardless of mount event, to register observer support."""
    # Call the original function being patched to ensure the widget is fully set up before registering.
    _widget_post_mount(self)
    attach_to_observers = getattr(self.app, "attach_to_observers", None)
    if attach_to_observers:
        attach_to_observers(self)
    if self.id:
        _register_reactive_observers = getattr(self.app, "_register_reactive_observers", None)
        if _register_reactive_observers:
            _register_reactive_observers(self)


# Patch the base Widget class to allow all reactive attributes to register to observer applications.
_widget_post_mount = Widget._post_mount  # pylint: disable=protected-access
Widget._post_mount = _post_mount_patch  # pylint: disable=protected-access



"""Widget metadata that is used to create an item within a ListView on demand."""

from typing import Any
from typing import Callable

from textual import events

from ._list_item import ListItem


class ListItemMeta:
    """A metadata class used to create, and recreate, list items that will be used in list views."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        item_type: type[ListItem] = ListItem,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        data: Any = None,
        styles: dict[str, Any] | None = None,
        disabled_messages: list[type[events.Message]] | None = None,
        callbacks: dict[str, Callable] | None = None,
    ) -> None:
        """Initialize metadata object to allow creating list items on demand.

        Args:
            item_type: Type of list item to instantiate with the metadata values.
            name: The name of the widget.
            id: The ID of the widget in the DOM.
            classes: The CSS classes for the widget.
            disabled: Whether the widget is disabled or not.
            data: Optional data associated with the list item.
                If no child is provided for display, the data will be searched for a "label" key to use in a Label.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        if not issubclass(item_type, ListItem):
            raise ValueError("Item type must be a subclass of ListItem")
        self.item_type = item_type
        self.name = name
        self.id = id
        self.classes = classes
        self.disabled = disabled
        self.data = data
        self.menu_index: int | None = None
        self.extension_configs = {
            "styles": styles,
            "disabled_messages": disabled_messages,
            "callbacks": callbacks,
        }

    def to_item(self) -> ListItem:
        """Create the underlying list item widget.

        Returns:
            Newly created item that can be used in a list view.
        """
        item = self.item_type(
            name=self.name,
            id=self.id,
            classes=self.classes,
            disabled=self.disabled,
            data=self.data,
            **self.extension_configs,
        )
        item.menu_index = self.menu_index
        return item



"""Modal screen to show a provided widget and add basic navigation."""

from typing import ClassVar

from textual.app import App
from textual.app import ComposeResult
from textual.binding import Binding
from textual.binding import BindingType
from textual.screen import ModalScreen
from textual.screen import ScreenResultCallbackType
from textual.screen import ScreenResultType
from textual.widget import AwaitMount
from textual.widget import Widget


class ModalDialog(ModalScreen):
    """Basic modal screen to show a provided widget and add basic navigation."""

    BINDINGS: ClassVar[list[BindingType]] = [
        Binding("escape", "dismiss()", show=False),
    ]
    DEFAULT_CSS = """
    ModalDialog {
        align: center middle;
        background: $surface 50%;
    }
    """

    def __init__(
        self,
        child: Widget,
        id: str | None = None,
        classes: str | None = None,
    ) -> None:
        """Initialize dialog content.

        Args:
            child: Widget to display when dialog is mounted.
            id: The ID of the dialog in the DOM.
            classes: The CSS classes for the dialog.
        """
        super().__init__(id=id, classes=classes)
        self.child = child

    def compose(self) -> ComposeResult:
        """Show the child provided on instantiation."""
        yield self.child

    def show(
        self,
        app: App,
        callback: ScreenResultCallbackType[ScreenResultType] | None = None,
    ) -> AwaitMount:
        """Display the dialog on top of an application.

        Alias for app.push_screen().

        Args:
            app: Application to push the dialog on top of.
            callback: An optional callback function that will be called if the screen is dismissed with a result.

        Returns:
            An optional awaitable that awaits the mounting of the screen and its children.
        """
        return app.push_screen(self, callback=callback)



"""Extended Textual Select widget."""

from typing import Any
from typing import Callable
from typing import Iterable

from textual import events
from textual import widgets
from textual.widgets._select import SelectType

from .._extensions import WidgetExtension


class Select(WidgetExtension, widgets.Select):
    """An extended widget to select from a list of possible options."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        options: Iterable[tuple[str, SelectType]],
        *,
        prompt: str = "Select",
        allow_blank: bool = True,
        value: SelectType | None = None,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        styles: dict[str, Any] | None = None,
        disabled_messages: list[type[events.Message]] | None = None,
        callbacks: dict[str, Callable] | None = None,
    ) -> None:
        """Initialize the Select control.

        Args:
            options: Options to select from.
            prompt: Text to show in the control when no option is select.
            allow_blank: Allow the selection of a blank option.
            value: Initial value (should be one of the values in `options`).
            name: The name of the select control.
            id: The ID of the control the DOM.
            classes: The CSS classes of the control.
            disabled: Whether the control is disabled or not.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        super().__init__(
            options,
            prompt=prompt,
            allow_blank=allow_blank,
            value=value,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self.__extend_widget__(
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )



"""Extended Textual Markdown widgets."""

from typing import Any
from typing import Callable

from textual import events
from textual import widgets
from textual.widgets._markdown import MarkdownIt

from .._extensions import WidgetExtension


class Markdown(WidgetExtension, widgets.Markdown):
    """An extended widget for rendering markdown."""

    def __init__(
        self,
        markdown: str | None = None,
        *,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        parser_factory: Callable[[], MarkdownIt] | None = None,
        styles: dict[str, Any] | None = None,
        disabled_messages: list[type[events.Message]] | None = None,
        callbacks: dict[str, Callable] | None = None,
    ) -> None:
        """Initialize a Markdown widget.

        Args:
            markdown: String containing Markdown or None to leave blank for now.
            name: The name of the widget.
            id: The ID of the widget in the DOM.
            classes: The CSS classes of the widget.
            parser_factory: A factory function to return a configured MarkdownIt instance.
                If "None`, a "gfm-like" parser is used.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        super().__init__(
            markdown=markdown,
            name=name,
            id=id,
            classes=classes,
            parser_factory=parser_factory,
        )
        self.__extend_widget__(
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )


class MarkdownViewer(WidgetExtension, widgets.MarkdownViewer):
    """An extended Markdown viewer widget."""

    def __init__(
        self,
        markdown: str | None = None,
        *,
        show_table_of_contents: bool = True,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        parser_factory: Callable[[], MarkdownIt] | None = None,
        styles: dict[str, Any] | None = None,
        disabled_messages: list[type[events.Message]] | None = None,
        callbacks: dict[str, Callable] | None = None,
    ) -> None:
        """Initialize a Markdown Viewer widget.

        Args:
            markdown: String containing Markdown, or None to leave blank.
            show_table_of_contents: Show a table of contents in a sidebar.
            name: The name of the widget.
            id: The ID of the widget in the DOM.
            classes: The CSS classes of the widget.
            parser_factory: A factory function to return a configured MarkdownIt instance.
                If "None", a "gfm-like" parser is used.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        super().__init__(
            markdown=markdown,
            show_table_of_contents=show_table_of_contents,
            name=name,
            id=id,
            classes=classes,
            parser_factory=parser_factory,
        )
        self.__extend_widget__(
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )



"""Custom textual Widgets extensions."""

import time
from inspect import isawaitable
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import Generator
from typing import Iterable

from rich.console import RenderableType
from rich.text import TextType
from textual import events
from textual.message import Message
from textual.reactive import reactive
from textual.widget import Widget


class Clickable:
    """Add reactive attributes for clicking on a widget.

    Use "update_n_clicks()" to trigger reactive attributes in a consistent manner.
    """

    # Number of times the widget has been clicked on.
    n_clicks: int = reactive(0, repaint=False, init=False)
    # Time (in milliseconds since 1970) since the last time n_clicks updated.
    n_clicks_timestamp: float = reactive(-1.0, repaint=False, init=False)
    # Disable n_clicks* properties.
    disable_n_clicks: bool = reactive(False, repaint=False, init=False)

    def update_n_clicks(self) -> None:
        """Update the number of times the widget has been clicked by one."""
        if not self.disable_n_clicks:
            self.n_clicks += 1

    def watch_n_clicks(self, _: int) -> None:
        """Monitor click count to update the last time clicked."""
        self.n_clicks_timestamp = time.time()


class WidgetExtension:
    """Extension for textual widgets to allow various overrides and control of behaviors at an instance level.

    Key Features:
    - Handles messages at lowest level available for control of disables, intercepts, and more per instance.
    - Allows declaring "on_*" functions for callbacks per instance during instantiation.
    - Allows declaring "style" attributes per instance during instantiation.

    Compared to disabling and intercepting messages at "post_message" level, this extension ensures that any
    messages added directly to the queue are handled based on Widget rules. Widgets that inherit from this
    class can also "intercept" the messages to route them manually, swap them with new messages, absorb them, etc.

    If a message is not disabled, or returned after being intercepted, it will then route to a local callback
    if one was provided to the widget on instantiation. Local callbacks provide additional isolation of message
    handling logic from the global app namespace. If there is no local callback for a message, or the local callback
    returns True (for continue), the message will be sent to the original widget message handler, and propagate
    throughout the app as normal.

    If creating a new basic widget, use WidgetInitExtension + Widget instead. If subclassing an existing,
    basic, textual widget,  the extension class must be placed before base class to ensure custom behaviors
    take priority. Example:
        class MyContainer(WidgetExtension, Container):
            ...
    """

    default_disabled_messages = ()

    def __extend_widget__(
        self,
        styles: dict[str, Any] | None = None,
        disabled_messages: Iterable[type[events.Message]] | None = None,
        callbacks: dict[str, Callable] | None = None,
    ) -> None:
        """Set up the widget's extensions.

        Args:
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
                Defaults to class default_disabled_messages attribute if None.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
                Callbacks behave the same as "on_*" functions declared at class level.
                Return "True" in callbacks to send the messages to the default widget handler.
                May be provided as direct arguments to "additional_configs". e.g. "on_button_pressed=on_press"
        """
        self._local_callback_map = {}

        if styles:
            self.__extend_widget_styles__(styles)

        if callbacks:
            self.__extend_widget_messaging_callbacks__(callbacks)

        # Allow messages to be disabled for this instance of the widget only, or use subclass defaults.
        if disabled_messages is None:
            disabled_messages = self.default_disabled_messages
        if disabled_messages:
            self.disable_messages(*disabled_messages)

    def __extend_widget_messaging_callbacks__(
        self,
        local_callbacks: dict[str, Callable],
    ) -> None:
        """Validate and set up message callbacks."""
        for key, value in local_callbacks.items():
            if not key.startswith("on_"):
                raise ValueError(
                    'Callbacks must start with "on_" and end with the name of the event type in camel_case'
                )
            self._local_callback_map[key] = value

    def __extend_widget_styles__(self, styles: dict) -> None:
        """Apply inline/local styles for the instance."""
        for key, value in styles.items():
            setattr(self.styles, key, value)

    def after(
        self,
        awaitable: Awaitable,
        func: Callable,
        *args: Any,
        **kwargs: Any,
    ) -> bool:
        """Await a coroutine's results and then run a function.

        Args:
            awaitable: Async result returned from a coroutine.
            func: The function to run after the awaitable has completed.
            args: Positional arguments to pass to the function run after awaitable completes.
            kwargs: Keyword arguments to pass to the function run after awaitable completes.

        Returns:
            True if the message was posted, False if this widget was closed / closing.
        """

        async def awaiter() -> None:
            """Coroutine to await an awaitable returned from another function, before running final function."""
            await awaitable
            result = func(*args, **kwargs)
            if isawaitable(result):
                await result

        return self.post_message(events.Callback(callback=awaiter))

    async def _handle_local_callback(self, message: Message) -> bool:
        """Route message to a local callback if available, or recommend sending to native widget message handler.

        If callback manually returns a truthy value, the message will also be handled by native widget message handler.
        """
        handle_natively = True
        handler_name = message.handler_name
        if handler_name in self._local_callback_map:
            handle_natively = self._local_callback_map[handler_name](message)
            if isawaitable(handle_natively):
                handle_natively = await handle_natively
        return handle_natively

    async def intercept_message(self, message: Message) -> Message | None:
        """Intercept a message for this widget before processing.

        Args:
            message: Original message pending processing.

        Returns:
            Message to process after being intercepted by this widget, None to disable further processing.
        """
        # Defaults to returning to the original message to process as normal.
        return message

    async def _on_message(self, message: Message) -> None:
        """Override default message processing to allow disables, intercepts, and local callbacks, at lowest level."""
        if not self.check_message_enabled(message):
            # Ensure no other handlers see the message as being valid for further processing.
            message.stop()
            message.prevent_default()
            return

        message = await self.intercept_message(message)
        if not message:
            return
        if not await self._handle_local_callback(message):
            message.stop()
            message.prevent_default()
        else:
            await super()._on_message(message)
        return None

    def walk_all_children(self) -> Generator[Widget, None, None]:
        """Walk the subtree rooted at this node, and return every descendant encountered.

        Compared to walk_children, this function will also walk all pending children. Pending children
        will be walked first, followed by results from walk_children().

        Yields:
            Every child, pending or standard, starting from the top down, and pending before standard.
        """
        for child in walk_all_children(self):
            yield child


class StaticInitExtension(WidgetExtension):
    """Extension for textual widgets that inherit from Static class.

    Includes all extensions provided by WidgetExtension, and automatically calls the extension set up.
    See WidgetExtension for full details on all extended features.

    Example:
        class Label(StaticInitExtension, widgets.Label):
            ...
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        renderable: RenderableType = "",
        *,
        expand: bool = False,
        shrink: bool = False,
        markup: bool = True,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        styles: dict[str, Any] | None = None,
        disabled_messages: Iterable[type[events.Message]] | None = None,
        callbacks: dict[str, Callable] | None = None,
    ) -> None:
        """Initialize a Static widget with extension arguments.

        Args:
            renderable: A Rich renderable, or string containing console markup.
            expand: Expand content if required to fill container.
            shrink: Shrink content if required to fill container.
            markup: True if markup should be parsed and rendered.
            name: Name of widget.
            id: ID of Widget.
            classes: Space separated list of class names.
            disabled: Whether the static is disabled or not.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        super().__init__(
            renderable,
            expand=expand,
            shrink=shrink,
            markup=markup,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self.__extend_widget__(
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )


class ToggleButtonInitExtension(WidgetExtension):
    """Extension for textual widgets that inherit from ToggleButton class.

    Includes all extensions provided by WidgetExtension, and automatically calls the extension set up.
    See WidgetExtension for full details on all extended features.

    Example:
        class Checkbox(ToggleButtonInitExtension, containers.Checkbox):
            ...
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        label: TextType = "",
        value: bool = False,
        button_first: bool = True,
        *,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        styles: dict[str, Any] | None = None,
        disabled_messages: Iterable[type[events.Message]] | None = None,
        callbacks: dict[str, Callable] | None = None,
    ) -> None:
        """Initialize the toggle.

        Args:
            label: The label for the toggle.
            value: The initial value of the toggle.
            button_first: Should the button come before the label, or after?
            name: The name of the toggle.
            id: The ID of the toggle in the DOM.
            classes: The CSS classes of the toggle.
            disabled: Whether the button is disabled or not.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        super().__init__(
            label=label,
            value=value,
            button_first=button_first,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self.__extend_widget__(
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )


class WidgetInitExtension(WidgetExtension):
    """Extension for textual widgets that inherit from base Widget class.

    Includes all extensions provided by WidgetExtension, and automatically calls the extension set up.
    See WidgetExtension for full details on all extended features.

    Example:
        class Container(WidgetInitExtension, containers.Container):
            ...
    """

    def __init__(
        self,
        *children: Widget,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        styles: dict[str, Any] | None = None,
        disabled_messages: Iterable[events.Message] | None = None,
        callbacks: dict[str, Callable] | None = None,
    ) -> None:
        """Initialize a Widget with support for extension arguments.

        Args:
            *children: Child widgets.
            name: The name of the widget.
            id: The ID of the widget in the DOM.
            classes: The CSS classes for the widget.
            disabled: Whether the widget is disabled or not.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        super().__init__(
            *children,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self.__extend_widget__(
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )


def walk_all_children(widget: Widget) -> Generator[Widget, None, None]:
    """Walk the subtree of a node, and return every descendant encountered.

    Compared to walk_children, this function will also walk all pending children. Pending children
    will be walked first, followed by results from walk_children().

    Yields:
        Every child, pending or standard, starting from the top down, and pending before standard.
    """
    for pending_child in getattr(widget, "_pending_children", []):
        yield pending_child
        for nested_child in walk_all_children(pending_child):
            yield nested_child
    for child in widget.walk_children():
        yield child



"""Extended Textual Collapsible widget."""

from typing import Any
from typing import Callable

from textual import events
from textual import widgets
from textual.widget import Widget

from .._extensions import WidgetExtension


class Collapsible(WidgetExtension, widgets.Collapsible):
    """An extended collapsible container."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        *children: Widget,
        title: str = "Toggle",
        collapsed: bool = True,
        collapsed_symbol: str = "▶",
        expanded_symbol: str = "▼",
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        styles: dict[str, Any] | None = None,
        disabled_messages: list[type[events.Message]] | None = None,
        callbacks: dict[str, Callable] | None = None,
    ) -> None:
        """Initialize the collapsible widget.

        Args:
            *children: Contents that will be collapsed/expanded.
            title: Title of the collapsed/expanded contents.
            collapsed: Default status of the contents.
            collapsed_symbol: Collapsed symbol before the title.
            expanded_symbol: Expanded symbol before the title.
            name: The name of the collapsible.
            id: The ID of the widget in the DOM.
            classes: The CSS classes of the widget.
            disabled: Whether the widget is disabled or not.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        super().__init__(
            *children,
            title=title,
            collapsed=collapsed,
            collapsed_symbol=collapsed_symbol,
            expanded_symbol=expanded_symbol,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self.__extend_widget__(
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )



"""Extended Textual widget to display simple static content, or use as a base class for more complex widgets."""

from textual import widgets

from .._extensions import StaticInitExtension


class Static(StaticInitExtension, widgets.Static):
    """An extended widget to display simple static content, or use as a base class for more complex widgets."""



"""Extended Textual, simple, clickable, button."""

from typing import Any
from typing import Callable

from rich.text import TextType
from textual import events
from textual import widgets
from textual.message import Message
from textual.widgets.button import ButtonVariant

from ._extensions import Clickable
from ._extensions import WidgetExtension


class Button(Clickable, WidgetExtension, widgets.Button):
    """An extended, simple, clickable button."""

    def __init__(
        self,
        label: TextType | None = None,
        variant: ButtonVariant = "default",
        *,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        styles: dict[str, Any] | None = None,
        disabled_messages: list[type[events.Message]] | None = None,
        callbacks: dict[str, Callable] | None = None,
    ) -> None:
        """Initialize a Button widget with extension arguments.

        Args:
            label: The text that appears within the button.
            variant: The variant of the button.
            name: The name of the button.
            id: The ID of the button in the DOM.
            classes: The CSS classes of the button.
            disabled: Whether the button is disabled or not.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        super().__init__(
            label=label,
            variant=variant,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self.__extend_widget__(
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )

    async def intercept_message(self, message: Message) -> Message | None:
        """Update reactive attributes on press."""
        if isinstance(message, Button.Pressed):
            self.update_n_clicks()
        return message
